import logging
import time
from typing import Any, Dict, Optional

import elasticapm

from assemblyline.common import forge
from assemblyline.common.constants import SERVICE_STATE_HASH, ServiceStatus
from assemblyline.common.dict_utils import flatten, unflatten
from assemblyline.common.heuristics import HeuristicHandler, InvalidHeuristicException
from assemblyline.common.isotime import now_as_iso
from assemblyline.common.threading import APMAwareThreadPoolExecutor
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.filestore import FileStore
from assemblyline.odm import construct_safe
from assemblyline.odm.messages.changes import Operation
from assemblyline.odm.messages.task import Task as ServiceTask
from assemblyline.odm.models.error import Error
from assemblyline.odm.models.heuristic import Heuristic
from assemblyline.odm.models.result import Result
from assemblyline.odm.models.service import Service
from assemblyline.odm.models.tagging import Tagging
from assemblyline.remote.datatypes.events import EventSender, EventWatcher
from assemblyline.remote.datatypes.hash import ExpiringHash
from assemblyline_core.dispatching.client import DispatchClient

class TaskingClientException(Exception):
    pass


class ServiceMissingException(Exception):
    pass


# Tasking class
class TaskingClient:
    """A helper class to simplify tasking for privileged services and service-server.

    This tool helps take care of interactions between the filestore,
    datastore, dispatcher, and any sources of files to be processed.
    """

    def __init__(self, datastore: AssemblylineDatastore = None, filestore: FileStore = None,
                 config=None, redis=None, redis_persist=None, identify=None, register_only=False):
        self.log = logging.getLogger('assemblyline.tasking_client')
        self.config = config or forge.CachedObject(forge.get_config)
        self.datastore = datastore or forge.get_datastore(self.config)
        self.event_sender = EventSender('changes', redis)
        self.cleanup = False
        self.event_listener = None

        # If we're performing service registration, we only need a connection to the datastore
        if not register_only:
            self.dispatch_client = DispatchClient(self.datastore, redis=redis, redis_persist=redis_persist)
            self.event_listener = EventWatcher(redis)
            self.filestore = filestore or forge.get_filestore(self.config)
            self.heuristic_handler = HeuristicHandler(self.datastore)
            self.heuristics: dict[str, Heuristic] = {}
            self.reload_heuristics({})
            self.status_table = ExpiringHash(SERVICE_STATE_HASH, ttl=60*30, host=redis)
            self.tag_safelister = forge.CachedObject(forge.get_tag_safelister, kwargs=dict(
                log=self.log, config=config, datastore=self.datastore), refresh=300)
            if identify:
                self.cleanup = False
            else:
                self.cleanup = True
            self.identify = identify or forge.get_identify(config=self.config, datastore=self.datastore, use_cache=True)
            self.event_listener.register('changes.heuristics', self.reload_heuristics)
            self.event_listener.start()

    def reload_heuristics(self, data: Optional[dict]):
        service_name = None
        if data is not None:
            service_name = data.get('service_name')

        if service_name:
            self.heuristics.update({h.heur_id: h for h in self.datastore.list_service_heuristics(service_name)})
        else:
            self.heuristics = {h.heur_id: h for h in self.datastore.list_all_heuristics()}

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.stop()

    def stop(self):
        if self.cleanup:
            self.identify.stop()
        if self.event_listener:
            self.event_listener.stop()

    @elasticapm.capture_span(span_type='tasking_client')
    def upload_file(self, file_path, classification, ttl, is_section_image, expected_sha256=None):
        # Identify the file info of the uploaded file
        file_info = self.identify.fileinfo(file_path)

        # Validate SHA256 of the uploaded file
        if expected_sha256 is None or expected_sha256 == file_info['sha256']:
            file_info['archive_ts'] = None
            file_info['classification'] = classification
            if ttl:
                file_info['expiry_ts'] = now_as_iso(ttl * 24 * 60 * 60)
            else:
                file_info['expiry_ts'] = None

            # Update the datastore with the uploaded file
            self.datastore.save_or_freshen_file(file_info['sha256'], file_info, file_info['expiry_ts'],
                                                file_info['classification'], is_section_image=is_section_image)

            # Upload file to the filestore (upload already checks if the file exists)
            self.filestore.upload(file_path, file_info['sha256'])
        else:
            raise TaskingClientException("Uploaded file does not match expected file hash. "
                                         f"[{file_info['sha256']} != {expected_sha256}]")

    # Service
    @elasticapm.capture_span(span_type='tasking_client')
    def register_service(self, service_data, log_prefix=""):
        keep_alive = True

        try:
            # Get heuristics list
            heuristics = service_data.pop('heuristics', None)

            # Patch update_channel, registry_type before Service registration object creation
            service_data['update_channel'] = service_data.get(
                'update_channel', self.config.services.preferred_update_channel)
            service_data['docker_config']['registry_type'] = service_data['docker_config'] \
                .get('registry_type', self.config.services.preferred_registry_type)
            service_data['privileged'] = service_data.get('privileged', self.config.services.prefer_service_privileged)
            for dep in service_data.get('dependencies', {}).values():
                dep['container']['registry_type'] = dep.get(
                    'registry_type', self.config.services.preferred_registry_type)

            # Pop unused registration service_data
            for x in ['file_required', 'tool_version']:
                service_data.pop(x, None)

            # Create Service registration object
            service = Service(service_data)

            # Fix service version, we don't need to see the stable label
            service.version = service.version.replace('stable', '')

            # Save service if it doesn't already exist
            if not self.datastore.service.exists(f'{service.name}_{service.version}'):
                self.datastore.service.save(f'{service.name}_{service.version}', service)
                self.datastore.service.commit()
                self.log.info(f"{log_prefix}{service.name} registered")
                keep_alive = False

            # Save service delta if it doesn't already exist
            if not self.datastore.service_delta.exists(service.name):
                self.datastore.service_delta.save(service.name, {'version': service.version})
                self.datastore.service_delta.commit()
                self.log.info(f"{log_prefix}{service.name} "
                              f"version ({service.version}) registered")

            new_heuristics = []
            if heuristics:
                plan = self.datastore.heuristic.get_bulk_plan()
                for index, heuristic in enumerate(heuristics):
                    heuristic_id = f'#{index}'  # Set heuristic id to it's position in the list for logging purposes
                    try:
                        # Append service name to heuristic ID
                        heuristic['heur_id'] = f"{service.name.upper()}.{str(heuristic['heur_id'])}"

                        # Attack_id field is now a list, make it a list if we receive otherwise
                        attack_id = heuristic.get('attack_id', None)
                        if isinstance(attack_id, str):
                            heuristic['attack_id'] = [attack_id]

                        heuristic = Heuristic(heuristic)
                        heuristic_id = heuristic.heur_id
                        plan.add_upsert_operation(heuristic_id, heuristic)
                    except Exception as e:
                        msg = f"{service.name} has an invalid heuristic ({heuristic_id}): {str(e)}"
                        self.log.exception(f"{log_prefix}{msg}")
                        raise ValueError(msg)

                for item in self.datastore.heuristic.bulk(plan)['items']:
                    if item['update']['result'] != "noop":
                        new_heuristics.append(item['update']['_id'])
                        self.log.info(f"{log_prefix}{service.name} "
                                      f"heuristic {item['update']['_id']}: {item['update']['result'].upper()}")

                self.datastore.heuristic.commit()

                # Notify components watching for heuristic config changes
                self.event_sender.send('heuristics', {
                    'operation': Operation.Modified,
                    'service_name': service.name
                })

            service_config = self.datastore.get_service_with_delta(service.name, as_obj=False)

            # Notify components watching for service config changes
            self.event_sender.send('services.' + service.name, {
                'operation': Operation.Added,
                'name': service.name
            })

        except ValueError as e:  # Catch errors when building Service or Heuristic model(s)
            raise e

        return dict(keep_alive=keep_alive, new_heuristics=new_heuristics, service_config=service_config or dict())

    # Task
    @elasticapm.capture_span(span_type='tasking_client')
    def get_task(self, client_id, service_name, service_version, service_tool_version,
                 metric_factory, status_expiry=None, timeout=30):
        if status_expiry is None:
            status_expiry = time.time() + timeout

        cache_found = False

        try:
            service_data = self.dispatch_client.service_data[service_name]
        except KeyError:
            raise ServiceMissingException("The service you're asking task for does not exist, try later", 404)

        # Set the service status to Idle since we will be waiting for a task
        self.status_table.set(client_id, (service_name, ServiceStatus.Idle, status_expiry))

        # Getting a new task
        task = self.dispatch_client.request_work(client_id, service_name, service_version, timeout=timeout)

        if not task:
            # We've reached the timeout and no task found in service queue
            return None, False

        # We've got a task to process, consider us busy
        self.status_table.set(client_id, (service_name, ServiceStatus.Running, time.time() + service_data.timeout))
        metric_factory.increment('execute')

        result_key = Result.help_build_key(sha256=task.fileinfo.sha256,
                                           service_name=service_name,
                                           service_version=service_version,
                                           service_tool_version=service_tool_version,
                                           is_empty=False,
                                           task=task)

        # If we are allowed, try to see if the result has been cached
        if not task.ignore_cache and not service_data.disable_cache:
            # Checking for previous results for this key
            result = self.datastore.result.get_if_exists(result_key)
            if result:
                metric_factory.increment('cache_hit')
                if result.result.score:
                    metric_factory.increment('scored')
                else:
                    metric_factory.increment('not_scored')

                result.archive_ts = None

                if task.ttl:
                    result.expiry_ts = now_as_iso(task.ttl * 24 * 60 * 60)

                # Create a list of files to freshen
                freshen_hashes = [task.fileinfo.sha256]

                # Test each extracted and supplementary files
                for file_item in result.response.extracted + result.response.supplementary:
                    if file_item.sha256 in freshen_hashes:
                        # We've already decided to freshen this file, moving on..
                        continue

                    freshen_hashes.append(file_item.sha256)

                    # Bail out if file does not exists
                    if not self.filestore.exists(file_item.sha256):
                        self.log.info("We have a cache hit with some related files missing, ignoring it...")
                        metric_factory.increment('cache_miss')
                        return task.as_primitives(), False

                # Freshen the files
                for sha256 in freshen_hashes:
                    self.datastore.save_or_freshen_file(sha256, {}, result.expiry_ts, result.classification)

                self.dispatch_client.service_finished(task.sid, result_key, result)
                cache_found = True

            if not cache_found:
                # Checking for previous empty results for this key
                try:
                    result = self.datastore.emptyresult.get_if_exists(f"{result_key}.e")
                except ValueError:
                    self.log.warning(f"Got poisoned empty result cache record for key {result_key}.e, cleaning up...")
                    self.datastore.emptyresult.delete(f"{result_key}.e")

                if result:
                    metric_factory.increment('cache_hit')
                    metric_factory.increment('not_scored')
                    result = self.datastore.create_empty_result_from_key(result_key)
                    self.dispatch_client.service_finished(task.sid, f"{result_key}.e", result)
                    cache_found = True

            if not cache_found:
                metric_factory.increment('cache_miss')

        else:
            metric_factory.increment('cache_skipped')

        if not cache_found:
            # No luck with the cache, lets dispatch the task to a client
            return task.as_primitives(), False

        return None, True

    @elasticapm.capture_span(span_type='tasking_client')
    def task_finished(self, service_task, client_id, service_name, metric_factory):
        exec_time = service_task.get('exec_time')

        try:
            task = ServiceTask(service_task['task'])

            if 'result' in service_task:  # Task created a result
                missing_files = self._handle_task_result(
                    exec_time, task, service_task['result'],
                    client_id, service_name, service_task['freshen'], metric_factory)
                if missing_files:
                    return dict(success=False, missing_files=missing_files)
                return dict(success=True)

            elif 'error' in service_task:  # Task created an error
                error = service_task['error']
                self._handle_task_error(exec_time, task, error, client_id, service_name, metric_factory)
                return dict(success=True)
            else:
                return None

        except ValueError as e:  # Catch errors when building Task or Result model
            raise e

    @elasticapm.capture_span(span_type='tasking_client')
    def _handle_task_result(self, exec_time: int, task: ServiceTask, result: Dict[str, Any], client_id, service_name,
                            freshen: bool, metric_factory):
        # In the event of a result with duplicate files, let's cache file existence checks with the filestore
        file_exists_check = {}

        def freshen_file(file_info_list, item):
            file_info = file_info_list.get(item['sha256'], None)
            if file_info is None or not file_exists_check[item['sha256']]:
                return True
            else:
                file_info['archive_ts'] = None
                file_info['expiry_ts'] = expiry_ts
                file_info['classification'] = item['classification']
                self.datastore.save_or_freshen_file(item['sha256'], file_info,
                                                    file_info['expiry_ts'], file_info['classification'],
                                                    is_section_image=item.get('is_section_image', False))
            return False

        if task.ttl:
            expiry_ts = now_as_iso(task.ttl * 24 * 60 * 60)
        else:
            expiry_ts = None

        # Check if all files are in the filestore
        if freshen:
            missing_files = []
            hashes = list(set([f['sha256']
                               for f in result['response']['extracted'] + result['response']['supplementary']]))
            # Pre-compute file existence checks before freshening files
            file_exists_check = {h: self.filestore.exists(h) for h in hashes}
            file_infos = self.datastore.file.multiget(hashes, as_obj=False, error_on_missing=False)

            with elasticapm.capture_span(name="handle_task_result.freshen_files",
                                         span_type="tasking_client"):
                with APMAwareThreadPoolExecutor(max_workers=5) as executor:
                    res = {
                        f['sha256']: executor.submit(freshen_file, file_infos, f)
                        for f in result['response']['extracted'] + result['response']['supplementary']}
                for k, v in res.items():
                    if v.result():
                        missing_files.append(k)

            if missing_files:
                return missing_files

        # Add scores to the heuristics, if any section set a heuristic
        with elasticapm.capture_span(name="handle_task_result.process_heuristics",
                                     span_type="tasking_client"):
            total_score = 0
            for section in result['result']['sections']:
                zeroize_on_sig_safe = section.pop('zeroize_on_sig_safe', True)
                section['tags'] = flatten(section['tags'])
                if section.get('heuristic'):
                    heur_id = f"{service_name.upper()}.{str(section['heuristic']['heur_id'])}"
                    section['heuristic']['heur_id'] = heur_id
                    try:
                        section['heuristic'], new_tags = self.heuristic_handler.service_heuristic_to_result_heuristic(
                            section['heuristic'], self.heuristics, zeroize_on_sig_safe)
                        for tag in new_tags:
                            section['tags'].setdefault(tag[0], [])
                            if tag[1] not in section['tags'][tag[0]]:
                                section['tags'][tag[0]].append(tag[1])
                        total_score += section['heuristic']['score']
                    except InvalidHeuristicException:
                        section['heuristic'] = None

        # Update the total score of the result
        result['result']['score'] = total_score

        # Add timestamps for creation, archive and expiry
        result['created'] = now_as_iso()
        result['archive_ts'] = None
        result['expiry_ts'] = expiry_ts

        # Pop the temporary submission data
        temp_submission_data = result.pop('temp_submission_data', None)
        if temp_submission_data:
            old_submission_data = {row.name: row.value for row in task.temporary_submission_data}
            temp_submission_data = {k: v for k, v in temp_submission_data.items()
                                    if k not in old_submission_data or v != old_submission_data[k]}
            big_temp_data = {k: len(str(v)) for k, v in temp_submission_data.items()
                             if len(str(v)) > self.config.submission.max_temp_data_length}
            if big_temp_data:
                big_data_sizes = [f"{k}={v}" for k, v in big_temp_data.items()]
                self.log.warning(f"[{task.sid}] The following temporary submission keys were ignored because they are "
                                 "bigger then the maximum data size allowed "
                                 f"[{self.config.submission.max_temp_data_length}]: {' | '.join(big_data_sizes)}")
                temp_submission_data = {k: v for k, v in temp_submission_data.items() if k not in big_temp_data}

        # Process the tag values
        with elasticapm.capture_span(name="handle_task_result.process_tags",
                                     span_type="tasking_client"):
            for section in result['result']['sections']:
                # Perform tag safelisting
                tags, safelisted_tags = self.tag_safelister.get_validated_tag_map(section['tags'])
                section['tags'] = unflatten(tags)
                section['safelisted_tags'] = safelisted_tags

                section['tags'], dropped = construct_safe(Tagging, section.get('tags', {}))

                # Set section score to zero and lower total score if service is set to zeroize score
                # and all tags were safelisted
                if section.pop('zeroize_on_tag_safe', False) and \
                        section.get('heuristic') and \
                        len(tags) == 0 and \
                        len(safelisted_tags) != 0:
                    result['result']['score'] -= section['heuristic']['score']
                    section['heuristic']['score'] = 0

                if dropped:
                    self.log.warning(f"[{task.sid}] Invalid tag data from {service_name}: {dropped}")

        result = Result(result)
        result_key = result.build_key(service_tool_version=result.response.service_tool_version, task=task)
        self.dispatch_client.service_finished(task.sid, result_key, result, temp_submission_data)

        # Metrics
        if result.result.score > 0:
            metric_factory.increment('scored')
        else:
            metric_factory.increment('not_scored')

        self.log.info(f"[{task.sid}] {client_id} - {service_name} "
                      f"successfully completed task {f' in {exec_time}ms' if exec_time else ''}")

        self.status_table.set(client_id, (service_name, ServiceStatus.Idle, time.time() + 5))

    @elasticapm.capture_span(span_type='tasking_client')
    def _handle_task_error(
            self, exec_time: int, task: ServiceTask, error: Dict[str, Any],
            client_id, service_name, metric_factory) -> None:
        self.log.info(f"[{task.sid}] {client_id} - {service_name} "
                      f"failed to complete task {f' in {exec_time}ms' if exec_time else ''}")

        # Add timestamps for creation, archive and expiry
        error['created'] = now_as_iso()
        error['archive_ts'] = None
        if task.ttl:
            error['expiry_ts'] = now_as_iso(task.ttl * 24 * 60 * 60)

        error = Error(error)
        error_key = error.build_key(service_tool_version=error.response.service_tool_version, task=task)
        self.dispatch_client.service_failed(task.sid, error_key, error)

        # Metrics
        if error.response.status == 'FAIL_RECOVERABLE':
            metric_factory.increment('fail_recoverable')
        else:
            metric_factory.increment('fail_nonrecoverable')

        self.status_table.set(client_id, (service_name, ServiceStatus.Idle, time.time() + 5))




import hashlib
import random
from copy import deepcopy

import pytest

from assemblyline.common.forge import get_classification
from assemblyline.common.isotime import iso_to_epoch, now_as_iso
from assemblyline.odm.random_data import create_users, create_safelists, wipe_users, wipe_safelist
from assemblyline.odm.randomizer import get_random_hash
from assemblyline_core.safelist_client import SafelistClient, InvalidSafehash

add_hash_file = "10" + get_random_hash(62)
add_error_hash = "11" + get_random_hash(62)
update_hash = "12" + get_random_hash(62)
update_conflict_hash = "13" + get_random_hash(62)
source_hash = "14" + get_random_hash(62)

CLASSIFICATION = get_classification()

NSRL_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "NSRL",
    "reason": [
        "Found as test.txt on default windows 10 CD",
        "Found as install.txt on default windows XP CD"
    ],
    "type": "external"}

NSRL2_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "NSRL2",
    "reason": [
        "File contains only AAAAs..."
    ],
    "type": "external"}

ADMIN_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "admin",
    "reason": [
        "Generates a lot of FPs",
    ],
    "type": "user"}

USER_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "user",
    "reason": [
        "I just feel like it!",
        "I just feel like it!",
    ],
    "type": "user"}


@pytest.fixture(scope="module")
def client(datastore_connection):
    try:
        create_users(datastore_connection)
        create_safelists(datastore_connection)
        yield SafelistClient(datastore_connection)
    finally:
        wipe_users(datastore_connection)
        wipe_safelist(datastore_connection)


# noinspection PyUnusedLocal
def test_safelist_add_file(client):
    # Generate a random safelist
    sl_data = {
        'hashes': {'md5': get_random_hash(32),
                   'sha1': get_random_hash(40),
                   'sha256': add_hash_file},
        'file': {'name': ['file.txt'],
                 'size': random.randint(128, 4096),
                 'type': 'document/text'},
        'sources': [NSRL_SOURCE, ADMIN_SOURCE],
        'type': 'file'
    }
    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == add_hash_file
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.safelist.get(add_hash_file, as_obj=False)

    # Test dates
    added = ds_sl.pop('added', None)
    updated = ds_sl.pop('updated', None)

    # File item will live forever
    assert ds_sl.pop('expiry_ts', None) is None

    assert added == updated
    assert added is not None and updated is not None

    # Make sure tag and signature are none
    tag = ds_sl.pop('tag', None)
    signature = ds_sl.pop('signature', None)
    assert tag is None
    assert signature is None

    # Test classification
    classification = ds_sl.pop('classification', None)
    assert classification is not None

    # Test enabled
    enabled = ds_sl.pop('enabled', None)
    assert enabled

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    # Test rest
    assert ds_sl == sl_data_original


def test_safelist_add_tag(client):
    tag_type = 'network.static.ip'
    tag_value = '127.0.0.1'
    hashed_value = f"{tag_type}: {tag_value}".encode('utf8')
    expected_qhash = hashlib.sha256(hashed_value).hexdigest()
    # Generate a random safelist
    sl_data = {
        'dtl': 15,
        'hashes': {'md5': hashlib.md5(hashed_value).hexdigest(),
                   'sha1': hashlib.sha1(hashed_value).hexdigest(),
                   'sha256': expected_qhash},
        'tag': {'type': tag_type,
                'value': tag_value},
        'sources': [NSRL_SOURCE, ADMIN_SOURCE],
        'type': 'tag'
    }
    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == expected_qhash
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.safelist.get(expected_qhash, as_obj=False)

    # Test dates
    added = ds_sl.pop('added', None)
    updated = ds_sl.pop('updated', None)

    # Tag item will live up to a certain date
    assert ds_sl.pop('expiry_ts', None) is not None

    assert added == updated
    assert added is not None and updated is not None

    # Make sure file and signature are None
    file = ds_sl.pop('file', {})
    signature = ds_sl.pop('signature', None)
    assert file is None
    assert signature is None

    # Test classification
    classification = ds_sl.pop('classification', None)
    assert classification is not None

    # Test enabled
    enabled = ds_sl.pop('enabled', None)
    assert enabled

    # Test rest, dtl should not exist anymore
    sl_data_original.pop('dtl', None)

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    # Test rest
    assert ds_sl == sl_data_original


def test_safelist_add_signature(client):
    sig_name = 'McAfee.Eicar'
    hashed_value = f"signature: {sig_name}".encode('utf8')
    expected_qhash = hashlib.sha256(hashed_value).hexdigest()

    # Generate a random safelist
    sl_data = {
        'hashes': {'md5': hashlib.md5(hashed_value).hexdigest(),
                   'sha1': hashlib.sha1(hashed_value).hexdigest(),
                   'sha256': expected_qhash},
        'signature': {'name': sig_name},
        'sources': [ADMIN_SOURCE],
        'type': 'signature'
    }

    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == expected_qhash
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.safelist.get(expected_qhash, as_obj=False)

    # Test dates
    added = ds_sl.pop('added', None)
    updated = ds_sl.pop('updated', None)

    # Signature item will live forever
    assert ds_sl.pop('expiry_ts', None) is None

    assert added == updated
    assert added is not None and updated is not None

    # Make sure file and signature are None
    file = ds_sl.pop('file', {})
    tag = ds_sl.pop('tag', None)
    assert file is None
    assert tag is None

    # Test classification
    classification = ds_sl.pop('classification', None)
    assert classification is not None

    # Test enabled
    enabled = ds_sl.pop('enabled', None)
    assert enabled

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    # Test rest
    assert ds_sl == sl_data_original


def test_safelist_add_invalid(client):
    # Generate a random safelist
    sl_data = {
        'hashes': {'sha256': add_error_hash},
        'sources': [USER_SOURCE],
        'type': 'file'}

    # Insert it and test return value
    with pytest.raises(ValueError) as conflict_exc:
        client.add_update(sl_data, user={"uname": "test"})

    assert 'for another user' in conflict_exc.value.args[0]


def test_safelist_update(client):
    cl_eng = get_classification()

    # Generate a random safelist
    sl_data = {
        'hashes': {'md5': get_random_hash(32),
                   'sha1': get_random_hash(40),
                   'sha256': update_hash},
        'file': {'name': [],
                 'size': random.randint(128, 4096),
                 'type': 'document/text'},
        'sources': [NSRL_SOURCE],
        'type': 'file'
    }

    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == update_hash
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.safelist.get(update_hash, as_obj=False)

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    # Test rest
    assert {k: v for k, v in ds_sl.items() if k in sl_data_original} == sl_data_original

    u_data = {
        'hashes': {'sha256': update_hash},
        'sources': [USER_SOURCE],
        'type': 'file'
    }

    # Insert it and test return value
    qhash, op = client.add_update(u_data)
    assert qhash == update_hash
    assert op == 'update'

    # Load inserted data from DB
    ds_u = client.datastore.safelist.get(update_hash, as_obj=False)

    # Normalize classification in sources
    for source in ds_u['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    assert ds_u['added'] == ds_sl['added']
    assert iso_to_epoch(ds_u['updated']) > iso_to_epoch(ds_sl['updated'])
    assert len(ds_u['sources']) == 2
    assert USER_SOURCE in ds_u['sources']
    assert NSRL_SOURCE in ds_u['sources']


def test_safelist_update_conflict(client):
    # Generate a random safelist
    sl_data = {'hashes': {'sha256': update_conflict_hash}, 'file': {}, 'sources': [ADMIN_SOURCE], 'type': 'file'}

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == update_conflict_hash
    assert op == 'add'

    # Insert the same source with a different type
    sl_data['sources'][0]['type'] = 'external'
    with pytest.raises(InvalidSafehash) as conflict_exc:
        client.add_update(sl_data)

    assert 'has a type conflict:' in conflict_exc.value.args[0]



import json
import requests
import re
import socket
import string

from assemblyline.common.version import FRAMEWORK_VERSION, SYSTEM_VERSION
from assemblyline.odm.models.config import Config as SystemConfig
from assemblyline.odm.models.service import Service as ServiceConfig

from base64 import b64encode
from collections import defaultdict
from logging import Logger
from packaging.version import parse, Version

DEFAULT_DOCKER_REGISTRY = "registry.hub.docker.com"


class ContainerRegistry():
    # Provide a means of obtaining a list of tags from a container registry
    def _get_proprietary_registry_tags(self, server, image_name, auth, verify, proxies=None, token_server=None):
        raise NotImplementedError()


class DockerRegistry(ContainerRegistry):
    def _get_proprietary_registry_tags(self, server, image_name, auth, verify, proxies=None, token_server=None):
        # Find latest tag for each types
        url = f"https://{server}/v2/{image_name}/tags/list"

        # Get tag list
        headers = {}
        if auth:
            headers["Authorization"] = auth
        else:
            # Retrieve token for authentication: https://distribution.github.io/distribution/spec/auth/token/

            # Assume the token server is the same as the container image registry host if not explicitly set
            token_server = token_server if token_server else server
            token_url = f"https://{token_server}/token?scope=repository:{image_name}:pull"
            resp = requests.get(token_url)
            if resp.ok:
                # Request to obtain token was successful, set Authorization header for registry API
                token = resp.json().get('token')
                headers["Authorization"] = f"Bearer {token}"

        resp = None
        try:
            resp = requests.get(url, headers=headers, verify=verify, proxies=proxies)
        except requests.exceptions.SSLError:
            # Connect to insecure registry over HTTP (development only)
            if not verify:
                url = f"http://{server}/v2/{image_name}/tags/list"
                resp = requests.get(url, headers=headers, verify=verify, proxies=proxies)

        # Test for valid response
        if resp and resp.ok:
            # Test for positive list of tags
            resp_data = resp.json()
            return resp_data['tags'] or []
        return []


class HarborRegistry(ContainerRegistry):
    def _get_proprietary_registry_tags(self, server, image_name, auth, verify, proxies=None, token_server=None):
        # Determine project/repo IDs from image name
        project_id, repo_id = image_name.split('/', 1)
        repo_id = repo_id.replace('/', "%2F")
        url = f"https://{server}/api/v2.0/projects/{project_id}/repositories/{repo_id}/artifacts?page_size=0"

        headers = {}
        if auth:
            headers["Authorization"] = auth
        else:
            # Retrieve token for authentication: https://github.com/goharbor/harbor/wiki/Harbor-FAQs#api

            # Assume the token server is the same as the container image registry host if not explicitly set
            token_server = token_server if token_server else server
            token_url = f"https://{server}/service/token?scope=repository:{image_name}:pull"
            resp = requests.get(token_url)
            if resp.ok:
                # Request to obtain token was successful, set Authorization header for registry API
                token = resp.json().get('token')
                headers["Authorization"] = f"Bearer {token}"
        resp = None
        try:
            resp = requests.get(url, headers=headers, verify=verify, proxies=proxies)
        except requests.exceptions.SSLError:
            # Connect to insecure registry over HTTP (development only)
            if not verify:
                url = f"http://{server}/api/v2.0/projects/{project_id}/repositories/{repo_id}/artifacts"
                resp = requests.get(url, headers=headers, verify=verify, proxies=proxies)

        if resp and resp.ok:
            return [tag['name'] for image in resp.json() if image['tags'] for tag in image['tags']]
        return []


REGISTRY_TYPE_MAPPING = {
    'docker': DockerRegistry(),
    'harbor': HarborRegistry()
}


def get_latest_tag_for_service(
        service_config: ServiceConfig, system_config: SystemConfig, logger: Logger, prefix: str = ""):
    def process_image(image):
        # Find which server to search in
        server = image.split("/")[0]
        if server != "cccs":
            if ":" in server:
                image_name = image[len(server) + 1:]
            else:
                try:
                    socket.gethostbyname_ex(server)
                    image_name = image[len(server) + 1:]
                except socket.gaierror:
                    server = DEFAULT_DOCKER_REGISTRY
                    image_name = image
        else:
            server = DEFAULT_DOCKER_REGISTRY
            image_name = image

        # Split repo name without the tag
        image_name = image_name.rsplit(":", 1)[0]

        return server, image_name

    # Extract info
    service_name = service_config.name
    image = service_config.docker_config.image
    update_channel = service_config.update_channel

    # Fix service image for calling Docker API
    image_variables = defaultdict(str)
    image_variables.update(system_config.services.image_variables)
    image_variables.update(system_config.services.update_image_variables)
    searchable_image = string.Template(image).safe_substitute(image_variables)

    # Get authentication
    auth = None
    auth_config = None
    server, image_name = process_image(searchable_image)

    if not (service_config.docker_config.registry_username and service_config.docker_config.registry_password):
        # If the passed in service configuration is missing registry credentials, check against system configuration
        for registry in system_config.services.registries:
            if server.startswith(registry['name']):
                # Apply the credentials that the system is configured to use with the registry
                service_config.docker_config.registry_username = registry['username']
                service_config.docker_config.registry_password = registry['password']
                service_config.docker_config.registry_type = registry['type']
                break

    if service_config.docker_config.registry_username and service_config.docker_config.registry_password:
        # We're authenticating using Basic Auth
        auth_config = {
            'username': service_config.docker_config.registry_username,
            'password': service_config.docker_config.registry_password
        }
        upass = f"{service_config.docker_config.registry_username}:{service_config.docker_config.registry_password}"
        auth = f"Basic {b64encode(upass.encode()).decode()}"
    elif service_config.docker_config.registry_password:
        # We're assuming that if only a password is given, then this is a token
        auth = f"Bearer {service_config.docker_config.registry_password}"

    registry = REGISTRY_TYPE_MAPPING[service_config.docker_config.registry_type]
    token_server = None
    proxies = None
    for reg_conf in system_config.core.updater.registry_configs:
        if reg_conf.name == server:
            proxies = reg_conf.proxies or None
            token_server = reg_conf.token_server or None
            break

    if server == DEFAULT_DOCKER_REGISTRY:
        tags = _get_dockerhub_tags(image_name, update_channel, proxies)
    else:
        tags = registry._get_proprietary_registry_tags(server, image_name, auth,
                                                       not system_config.services.allow_insecure_registry,
                                                       proxies, token_server)

    tag_name = None

    # Pre-filter tags to only consider 'compatible' tags relative to the running system
    tags = [t for t in tags
            if re.match(f"({FRAMEWORK_VERSION})[.]({SYSTEM_VERSION})[.]\\d+[.]({update_channel})\\d+", t)]

    if not tags:
        logger.warning(f"{prefix}Cannot fetch latest tag for service {service_name} - {image_name}"
                       f" => [server: {server}, repo_name: {image_name}, channel: {update_channel}]")
    else:
        for t in tags:
            t_version = Version(t.replace(update_channel, ""))
            # Tag name gets assigned to the first viable option then relies on comparison to get the latest
            if not tag_name:
                tag_name = t
            elif t_version.major == FRAMEWORK_VERSION and t_version.minor == SYSTEM_VERSION and \
                    t_version > parse(tag_name.replace(update_channel, "")):
                tag_name = t

        logger.info(f"{prefix}Latest {service_name} tag on {update_channel.upper()} channel is: {tag_name}")

    # Fix service image for use in Kubernetes
    image_variables = defaultdict(str)
    image_variables.update(system_config.services.image_variables)
    image = string.Template(image).safe_substitute(image_variables)
    server, image_name = process_image(image)

    # Append server to image if not the default server
    if server != "registry.hub.docker.com":
        image_name = "/".join([server, image_name])

    return image_name, tag_name, auth_config


# Default for obtaining tags from DockerHub
def _get_dockerhub_tags(image_name, update_channel, proxies=None):
    # Find latest tag for each types
    url = f"https://{DEFAULT_DOCKER_REGISTRY}/v2/repositories/{image_name}/tags" \
        f"?page_size=5&page=1&name={update_channel}"

    # Get tag list
    resp = requests.get(url, proxies=proxies)

    # Test for valid response
    if resp.ok:
        # Test for positive list of tags
        resp_data = resp.json()
        return [x['name'] for x in resp_data['results']]

    return []



import time
import threading
import functools
import logging
from typing import Optional
import json
from collections import namedtuple

import requests


logger = logging.getLogger("assemblyline.vacuum.stream_map")

Stream = namedtuple('Stream', [
    'id',
    'name',
    'description',
    'zone_id',
    'classification'
])


class StreamMap:
    UPDATE_INTERVAL = 60 * 15

    @staticmethod
    @functools.cache
    def load(url: Optional[str], init: Optional[str]):
        return StreamMap(url, init)

    def __init__(self, url: Optional[str], init: Optional[str]):
        self.url = url
        self.init_data = init
        self.lock = threading.Lock()
        self.table: dict[int, Stream] = {}
        self.update_time = 0
        self._load_stream_map()

    def _load_stream_map(self):
        # Don't load more than once every 5 seconds
        if time.time() - self.update_time < 5:
            return

        with self.lock:
            # Recheck in case it was updated while waiting for lock
            if time.time() - self.update_time < 5:
                return

            table = {}
            try:
                if self.init_data:
                    for stream in json.loads(self.init_data):
                        stream = Stream(
                            id=int(stream['STREAM_ID']),
                            name=stream['STREAM_NAME'],
                            description=stream['STREAM_DESCRIPTION'],
                            zone_id=stream['ZONE'],
                            classification=f"{stream.get('LEVEL', 'PB')}//{stream.get('CAVEAT', 'CND')}"
                        )
                        table[stream.id] = stream

            except Exception:
                logger.exception("Error parsing stream_map_init data")

            try:
                if self.url:
                    res = requests.get(self.url, verify=False)
                    res.raise_for_status()

                    for stream in res.json()['data']:
                        stream = Stream(
                            id=int(stream['STREAM_ID']),
                            name=stream['STREAM_NAME'],
                            description=stream['STREAM_DESCRIPTION'],
                            zone_id=stream['ZONE'],
                            classification=f"{stream.get('LEVEL', 'PB')}//{stream.get('CAVEAT', 'CND')}"
                        )
                        table[stream.id] = stream
            except Exception:
                logger.exception("Error parsing stream_map_url data")

            self.table = table
            self.update_time = time.time()

    def _refresh_stream_map(self):
        if time.time() - self.update_time > self.UPDATE_INTERVAL:
            self._load_stream_map()

    def __getitem__(self, stream_id: int) -> Optional[Stream]:
        self._refresh_stream_map()
        try:
            return self.table[stream_id]
        except KeyError:
            self._load_stream_map()
            return self.table.get(stream_id)


if __name__ == '__main__':
    streams = StreamMap('', None)
    print(streams[10])
    print(streams[10000000])
